/*
    Copyright 2008,2009 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl-2.0.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <zlib.h>

typedef uint8_t     u8;
typedef uint16_t    u16;
typedef uint32_t    u32;



#define VER         "0.1.1"



u8 *newleecher_dec(u8 *buff, int *buffsz);
u8 *base64_decode(u8 *data, int *size);
u8 *fdload(u8 *fname, int *fsize);
void write_file(u8 *fname, u8 *buff, int buffsz);
void std_err(void);



int main(int argc, char *argv[]) {
    int     buffsz;
    u8      *buff;

    fputs("\n"
        "NewsLeecher files decrypter "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stdout);

    if(argc < 3) {
        printf("\n"
            "Usage: %s <file.DAT/BAK> <output_file>\n"
            "\n"
            "the encrypted files are located in the following folder and its subfolders:\n"
            "  %%appdata%%\\NewsLeecher\n"
            "\n", argv[0]);
        exit(1);
    }

    buff = fdload(argv[1], &buffsz);

    printf("- loaded %d bytes\n", buffsz);
    buff = newleecher_dec(buff, &buffsz);
    printf("- decrypted/decompressed %d bytes\n", buffsz);

    write_file(argv[2], buff, buffsz);

    free(buff);
    printf("- done\n");
    return(0);
}



u8 *newleecher_dec(u8 *buff, int *buffsz) {
    uLongf  tmpsz;
    int     i,
            dobase64;
    u16     bx;
    u8      c,
            *tmp,
            *line,
            *limit,
            *p,
            *l,
            *out;

    bx = 0x215f;
    for(i = 0; i < *buffsz; i++) {
        c = buff[i];
        buff[i] ^= bx >> 8;
        bx = ((bx + c) * 0xce6d) + 0x58bf;
    }

    tmpsz = *buffsz * 12;   // enough?
    tmp = malloc(tmpsz);
    if(!tmp) std_err();
    if(uncompress(tmp, &tmpsz, buff, *buffsz) != Z_OK) {
        printf("\n"
            "Error: problems during the decompression of the data\n"
            "       the input file is probably already in plain-text\n");
        exit(1);
    }

    dobase64 = 1;
    for(i = 0; i < tmpsz; i++) {    // simple method
        c = tmp[i];
        if((c == '\n') || (c == '\r')) continue;
        if(c < ' ') {
            dobase64 = 0;
            break;
        }
    }

    out = tmp + tmpsz;
    if(dobase64) {
        out = tmp;
        limit = tmp + tmpsz;
        for(p = tmp; p < limit; p = l) {
            for(l = p; (l < limit) && (*l != '\r') && (*l != '\n'); l++);
            *l = 0;
            for(l++; (l < limit) && ((*l == '\r') || (*l == '\n')); l++);
            line = base64_decode(p, NULL);
            out += sprintf(out, "%s\r\n", line);    // output is shorter than input, so I reuse the same buffer
            free(line);
        }
    }

    *buffsz = out - tmp;
    return(tmp);
}



u8 *base64_decode(u8 *data, int *size) {
    int     len,
            xlen,
            a   = 0,
            b   = 0,
            c,
            step;
    u8      *buff,
            *limit,
            *p;
    static const u8 base[128] = {   // supports also the Gamespy base64
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x3e,0x00,0x00,0x00,0x3f,
        0x34,0x35,0x36,0x37,0x38,0x39,0x3a,0x3b,0x3c,0x3d,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0a,0x0b,0x0c,0x0d,0x0e,
        0x0f,0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17,0x18,0x19,0x3e,0x00,0x3f,0x00,0x00,
        0x00,0x1a,0x1b,0x1c,0x1d,0x1e,0x1f,0x20,0x21,0x22,0x23,0x24,0x25,0x26,0x27,0x28,
        0x29,0x2a,0x2b,0x2c,0x2d,0x2e,0x2f,0x30,0x31,0x32,0x33,0x00,0x00,0x00,0x00,0x00
    };

    if(!size || (*size < 0)) {      // use size -1 for auto size!
        len = strlen(data);
    } else {
        len = *size;
    }
    xlen = ((len >> 2) * 3) + 1;    // NULL included in output for text
    buff = malloc(xlen);
    if(!buff) return(NULL);

    p = buff;
    limit = data + len;

    for(step = 0; /* data < limit */; step++) {
        do {
            if(data >= limit) {
                c = 0;
                break;
            }
            c = *data;
            data++;
            if((c == '=') || (c == '_')) {  // supports also the Gamespy base64
                c = 0;
                break;
            }
        } while(c && ((c <= ' ') || (c > 0x7f)));
        if(!c) break;

        switch(step & 3) {
            case 0: {
                a = base[c];
                break;
            }
            case 1: {
                b = base[c];
                *p++ = (a << 2)        | (b >> 4);
                break;
            }
            case 2: {
                a = base[c];
                *p++ = ((b & 15) << 4) | (a >> 2);
                break;
            }
            case 3: {
                *p++ = ((a & 3) << 6)  | base[c];
                break;
            }
        }
    }
    *p = 0;

    len = p - buff;
    if(size) *size = len;
    if((len + 1) != xlen) {             // save memory
        buff = realloc(buff, len + 1);  // NULL included!
        if(!buff) return(NULL);
    }
    return(buff);
}



u8 *fdload(u8 *fname, int *fsize) {
    struct stat xstat;
    FILE    *fd;
    int     size;
    u8      *buff;

    printf("- load %s\n", fname);
    fd = fopen(fname, "rb");
    if(!fd) std_err();
    fstat(fileno(fd), &xstat);
    size = xstat.st_size;
    buff = malloc(size);
    if(!buff) std_err();
    fread(buff, 1, size, fd);
    fclose(fd);
    *fsize = size;
    return(buff);
}



void write_file(u8 *fname, u8 *buff, int buffsz) {
    FILE    *fd;
    u8      ans[16];

    printf("- write %s\n", fname);
    fd = fopen(fname, "rb");
    if(fd) {
        fclose(fd);
        printf("- the file %s already exists\n  do you want to overwrite it (y/N)?  ", fname);
        fgets(ans, sizeof(ans), stdin);
        if(tolower(ans[0]) != 'y') exit(1);
    }
    fd = fopen(fname, "wb");
    if(!fd) std_err();
    fwrite(buff, 1, buffsz, fd);
    fclose(fd);
}



void std_err(void) {
    perror("\nError");
    exit(1);
}


